/****************************************************************************
**  SCALASCA    http://www.scalasca.org/                                   **
*****************************************************************************
**  Copyright (c) 1998-2022                                                **
**  Forschungszentrum Juelich GmbH, Juelich Supercomputing Centre          **
**                                                                         **
**  Copyright (c) 2009-2013                                                **
**  German Research School for Simulation Sciences GmbH,                   **
**  Laboratory for Parallel Programming                                    **
**                                                                         **
**  This software may be modified and distributed under the terms of       **
**  a BSD-style license.  See the COPYING file in the package base         **
**  directory for details.                                                 **
****************************************************************************/


#ifndef PEARL_EVENT_REP_H
#define PEARL_EVENT_REP_H


#include <iosfwd>

#include <pearl/ScopedEnum.h>
#include <pearl/SmallObject.h>
#include <pearl/pearl_padding.h>
#include <pearl/pearl_types.h>


/*-------------------------------------------------------------------------*/
/**
 *  @file    Event_rep.h
 *  @ingroup PEARL_base
 *  @brief   Declaration of the class Event_rep.
 *
 *  This header file provides the declaration of the class Event_rep.
 **/
/*-------------------------------------------------------------------------*/


namespace pearl
{
// --- Forward declarations -------------------------------------------------

class Buffer;
class Callsite;
class GlobalDefs;
class LocalTrace;
class MpiComm;
class MpiGroup;
class MpiWindow;
class Region;


/*-------------------------------------------------------------------------*/
/**
 *  @brief   Helper struct defining collective operation type enumerators.
 *  @ingroup PEARL_base
 *
 *  The CollectiveOpType_enum helper struct defines the enumerators provided
 *  by the scoped enumeration CollectiveOpType.  It is passed as a base class
 *  to the ScopedEnum class template.
 **/
/*-------------------------------------------------------------------------*/

struct CollectiveOpType_enum
{
    /// @brief Collective operation types.
    ///
    /// The scoped enumeration CollectiveOpType provides symbolic names for
    /// different types of collective operations.  These are used by a
    /// number of collective event representations to specify the measured
    /// operation in more detail.
    ///
    enum type
    {
        ALLGATHER,                     ///< Gather data from all to all group members (same count)
        ALLGATHERV,                    ///< Gather data from all to all group members (varying count)
        ALLOCATE,                      ///< Memory allocation
        ALLREDUCE,                     ///< Reduction from all to all group members
        ALLTOALL,                      ///< Combined scatter/gather operation (same count)
        ALLTOALLV,                     ///< Combined scatter/gather operation (varying count)
        ALLTOALLW,                     ///< Combined scatter/gather operation (varying count and data type)
        BARRIER,                       ///< Barrier synchronization
        BCAST,                         ///< Broadcast data from one to all group members
        CREATE_HANDLE,                 ///< Handle creation
        CREATE_HANDLE_AND_ALLOCATE,    ///< Combined handle creation and memory allocation
        DEALLOCATE,                    ///< Memory deallocation
        DESTROY_HANDLE,                ///< Handle destruction
        DESTROY_HANDLE_AND_DEALLOCATE, ///< Combined handle destruction and memory deallocation
        EXSCAN,                        ///< Prefix reduction excluding receiving group member
        GATHER,                        ///< Gather data from all to one group member (same count)
        GATHERV,                       ///< Gather data from all to one group member (varying count)
        REDUCE,                        ///< Reduction from all to one group member
        REDUCE_SCATTER,                ///< Combined reduce/scatter operation (varying count)
        REDUCE_SCATTER_BLOCK,          ///< Combined reduce/scatter operation (same count)
        SCAN,                          ///< Prefix reduction including receiving group member
        SCATTER,                       ///< Scatter data from one to all group members (same count)
        SCATTERV                       ///< Scatter data from one to all group members (varying count)
    };
};


/// @copydoc CollectiveOpType_enum::type
/// See CollectiveOpType_enum::type for details.
typedef ScopedEnum< CollectiveOpType_enum > CollectiveOpType;


/*-------------------------------------------------------------------------*/
/**
 *  @class   Event_rep
 *  @ingroup PEARL_base
 *  @brief   Base class for all event representations.
 **/
/*-------------------------------------------------------------------------*/

class PEARL_NOPAD_ATTR Event_rep
    : public SmallObject
{
    public:
        /// @name Constructors & destructor
        /// @{

        virtual
        ~Event_rep();

        /// @}
        /// @name Event type information
        /// @{

        virtual event_t
        getType() const = 0;

        virtual bool
        isOfType(event_t type) const = 0;

        /// @}
        /// @name Comparison operators
        /// @{

        bool
        operator==(const timestamp_t& rhs) const;

        bool
        operator!=(const timestamp_t& rhs) const;

        bool
        operator<(const timestamp_t& rhs) const;

        bool
        operator>(const timestamp_t& rhs) const;

        /// @}
        /// @name Access event data
        /// @{

        timestamp_t
        getTimestamp() const;

        virtual const Region&
        getRegion() const;

        virtual const Callsite&
        getCallsite() const;

        virtual MpiComm*
        getComm() const;

        virtual uint32_t
        getRoot() const;

        virtual uint64_t
        getBytesSent() const;

        virtual uint64_t
        getBytesReceived() const;

        virtual uint32_t
        getSource() const;

        virtual uint32_t
        getDestination() const;

        virtual uint32_t
        getTag() const;

        virtual uint64_t
        getRequestId() const;

        virtual uint32_t
        getLockId() const;

        virtual MpiWindow*
        get_window() const;

        virtual uint32_t
        get_remote() const;

        virtual uint32_t
        get_rma_id() const;

        virtual MpiGroup*
        get_group() const;

        virtual bool
        is_sync() const;

        virtual bool
        is_exclusive() const;

        #ifdef PEARL_ENABLE_METRICS
            virtual uint64_t
            getMetric(int index) const;
        #endif

        /// @}
        /// @name Modify event data
        /// @{

        void
        setTimestamp(timestamp_t timestamp);

        virtual void
        setRequestId(uint64_t requestId);

        /// @}


    protected:
        /// @name Constructors & destructor
        /// @{

        Event_rep(pearl::timestamp_t timestamp);

        Event_rep(const GlobalDefs& defs,
                  Buffer&           buffer);

        /// @}
        /// @name Serialize event data
        /// @{

        virtual void
        pack(Buffer& buffer) const;

        /// @}
        /// @name Generate human-readable output of event data
        /// @{

        virtual std::ostream&
        output(std::ostream& stream) const;

        /// @}
        /// @name Find previous/next communication request entries
        /// @{

        virtual uint32_t
        get_next_reqoffs() const;

        virtual uint32_t
        get_prev_reqoffs() const;

        virtual void
        set_prev_reqoffs(uint32_t offset);

        virtual void
        set_next_reqoffs(uint32_t offset);

        /// @}


    private:
        /// Timestamp when the event has occured
        timestamp_t mTimestamp;


        /* Declare friends */
        friend class Buffer;
        friend class Event;
        friend class RemoteEvent;

        friend void
        PEARL_preprocess_trace(const GlobalDefs& defs,
                               const LocalTrace& trace);
};


// --- Related functions ----------------------------------------------------

template< class EvtRepT >
inline EvtRepT&
event_cast(Event_rep& event)
{
    #if defined(HAVE_SCALASCA_DEBUG)
        return dynamic_cast< EvtRepT& >(event);
    #else
        return static_cast< EvtRepT& >(event);
    #endif
}


/// @name Stream I/O functions
/// @{

/// @brief Stream output operator.
///
/// Prints the given collective operation enumeration value @p item to an
/// output @p stream in a human-readable form.
///
/// @param stream
///     Output stream
/// @param item
///     Collective operation enumeration value
/// @returns
///     Output stream
///
std::ostream&
operator<<(std::ostream&    stream,
           CollectiveOpType item);

/// @}


// --- Inline methods -------------------------------------------------------

#include "Inline/Event_rep-inl.h"
}    // namespace pearl


#endif    // !PEARL_EVENT_REP_H
